package com.dvag.abschluss.quartz.observability

import com.dvag.example.otel.quartz.observability.QuartzInstrumentationJobListener.Companion.TRACE_ID
import io.opentelemetry.api.GlobalOpenTelemetry
import io.opentelemetry.api.trace.Span
import io.opentelemetry.api.trace.SpanContext
import io.opentelemetry.api.trace.TraceFlags
import io.opentelemetry.api.trace.TraceState
import io.opentelemetry.context.Context
import io.opentelemetry.context.Scope
import io.opentelemetry.instrumentation.api.instrumenter.Instrumenter
import io.opentelemetry.instrumentation.api.instrumenter.SpanKindExtractor
import io.opentelemetry.sdk.trace.IdGenerator
import org.quartz.JobDataMap
import org.quartz.JobExecutionContext
import org.quartz.JobExecutionException
import org.springframework.stereotype.Service

/**
 * Instrumentation for Quartz jobs.
 *
 * This class is responsible for starting and ending spans for Quartz jobs.
 * It uses the <code>OpenTelemetry Instrumenter API</code> to create spans.
 * We need our own instrumentation because the OpenTelemetry Quartz instrumentation
 * has some weird behavior.
 *
 * Improvements, compared to the OpenTelemetry Quartz instrumentation:
 * - The trace ID is taken from the job data map, so it can be propagated.
 * - The span name is the job class name.
 * - The error cause is extracted from the job exception. RetryableJob exceptions are handled correctly.
 *
 * Note: the OpenTelemetry Quartz instrumentation should be disabled to avoid duplicates.
 * This can be done via environment variable:
 * <code>
 * OTEL_INSTRUMENTATION_QUARTZ_ENABLED=false
 * </code>
 */
class QuartzTelemetryInstrumentation(
    private val instrumenter: Instrumenter<JobExecutionContext, Void>,
) {
    constructor() : this(
        Instrumenter
            .builder<JobExecutionContext, Void>(GlobalOpenTelemetry.get(), "quartz-custom", QuartzSpanNameExtractor())
            .setErrorCauseExtractor(QuartzErrorCauseExtractor())
            .addAttributesExtractor(QuartzAttributesExtractor())
            .buildInstrumenter(SpanKindExtractor.alwaysServer()),
    )

    private val otelContext: ThreadLocal<Context> = ThreadLocal()
    private val scope: ThreadLocal<Scope> = ThreadLocal()

    fun beforeJob(context: JobExecutionContext) {
        val spanContext =
            SpanContext.createFromRemoteParent(
                getParentTraceId(context.jobDetail.jobDataMap),
                IdGenerator.random().generateSpanId(),
                TraceFlags.getSampled(),
                TraceState.getDefault(),
            )
        val parentCtx = Context.root().with(Span.wrap(spanContext))
        if (instrumenter.shouldStart(parentCtx, context)) {
            val otelContext = instrumenter.start(parentCtx, context)
            this.otelContext.set(otelContext)
            this.scope.set(otelContext.makeCurrent())
        }
    }

    fun afterJob(context: JobExecutionContext?, jobException: JobExecutionException?) {
        this.scope.get()?.close()
        val otelContext = this.otelContext.get()
        this.otelContext.remove()
        if (otelContext != null) {
            instrumenter.end(otelContext, context!!, null, jobException)
        }
    }

    /**
     * Extracts the trace ID from the job data map.
     * This way we propagate the parent trace ID to the Quartz job.
     * If no trace ID is found, a new one is generated.
     */
    private fun getParentTraceId(jobDataMap: JobDataMap): String = jobDataMap.getString(TRACE_ID) ?: IdGenerator.random().generateTraceId()
}
